/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
/*
 * Functions and data relating to the file system browser.
 *
 * The functions are sorted alphabetically by name.
 *
 * When we're in the file system browser, information about each directory
 * entry in our current working directory is saved in a `struct dir_list_entry'
 * and is part of a linked list whose first node is pointed to by
 * `first_dir_list_entry' and whose last node `last_dir_list_entry'.
 * We don't save the full path of the directory entries in
 * `struct dir_list_entry', only the basename, so we must always xchdir() to
 * the directory the user sees in the browser. The reason we xchdir() is that
 * we keep track of the current working directory ourselves, because then we
 * don't have to bother checking return values of getcwd() and its family.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include "main.h"

#include "browser.h"
#include "curs_main.h"
#include "curs_menu.h"
#include "curs_lib.h"
#include "keymap.h"
#include "lib.h"
#include "misc.h"
#include "options.h"
#include "playlist.h"

struct dir_list_entry {
    char *menu_line;		  /* line displayed in the menu */
    char *basename;               /* basename of the dir entry */
    mode_t lstat_st_mode;	  /* st_mode member of the stat struct
				     filled in by lstat() */
    struct dir_list_entry *prev;  /* pointer to previous entry */
    struct dir_list_entry *next;  /* pointer to next entry */
};

static struct dir_list_entry *first_dir_list_entry;
static struct dir_list_entry *last_dir_list_entry;

static void add_dir_entry(const char *basename, const struct stat *s);
static struct dir_list_entry *get_pointer_to_menu_entry(int index);

/*
 * If entry number `index' in the directory listing is a directory or a symlink
 * to one, add all registered file types in it to the end of the playlist.
 * Return number of songs added (>=0) on success, otherwise tell the user
 * about the error and return -1.
 */
int browser_add_dir(int index)
{
    struct stat st;
    char path[PATH_MAX + 1];
    struct dir_list_entry *p;

    p = get_pointer_to_menu_entry(index);

    if (!S_ISDIR(p->lstat_st_mode) && !S_ISLNK(p->lstat_st_mode)) {
	msg("%s is not a directory.", p->basename);
	return -1;
    }

    /* See if the link points to a directory. */
    if (S_ISLNK(p->lstat_st_mode)) {
	if (stat(p->basename, &st) < 0) {
	    perror_msg("Can't stat %s", p->basename);
	    return -1;
	}
	if (!S_ISDIR(st.st_mode)) {
	    msg("Symlink %s does not point to a directory.",
		     p->basename);
	    return -1;
	}
    }

    /* add_dir() wants absolute paths */
    snprintf(path, sizeof(path), "%s%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/",
	     p->basename);

    /* add_dir() returns number of songs added (>=0) or -1 on error. */
    return add_dir(path);
}

/*
 * Add directory entry to the end of the directory listing using the
 * information from `basename' and `st', which must have been filled in
 * by lstat().
 */
static void add_dir_entry(const char *basename, const struct stat *st)
{
    char menu_line[STRING];
    struct dir_list_entry *new_entry;
    /* Max eight chars are used for these fields in the directory listing. */
    char owner[8+1], group[8+1];
    struct passwd *pw;
    struct group *gr;

    pw = getpwuid(st->st_uid);
    gr = getgrgid(st->st_gid);
    /* Resolve names if possible, else use ids. */
    pw ? strfcpy(owner, pw->pw_name, sizeof(owner))
	: snprintf(owner, sizeof(owner), "%d", (int) st->st_uid);
    gr ? strfcpy(group, gr->gr_name, sizeof(group))
	: snprintf(group, sizeof(group), "%d", (int) st->st_gid);

    snprintf(menu_line, sizeof(menu_line), "%c%c%c%c%c%c%c%c%c%c "
	     "%2d %-8.8s %-8.8s %8ld %s%s",
	     S_ISREG(st->st_mode) ? '-' : S_ISDIR(st->st_mode) ? 'd' :
	     S_ISLNK(st->st_mode) ? 'l' : S_ISBLK(st->st_mode) ? 'b' :
	     S_ISCHR(st->st_mode) ? 'c' : S_ISFIFO(st->st_mode) ? 'p' :
	     S_ISSOCK(st->st_mode) ? 's' : '-', /* '-' if unknown */
	     
	     (st->st_mode & S_IRUSR) != 0 ? 'r' : '-',
	     (st->st_mode & S_IWUSR) != 0 ? 'w' : '-',
	     (st->st_mode & S_ISUID) != 0 ? 's' :
	     (st->st_mode & S_IXUSR) ? 'x' : '-',

	     (st->st_mode & S_IRGRP) != 0 ? 'r' : '-',
	     (st->st_mode & S_IWGRP) != 0 ? 'w' : '-',
	     (st->st_mode & S_ISGID) != 0 ? 's' :
	     (st->st_mode & S_IXGRP) ? 'x' : '-',

	     (st->st_mode & S_IROTH) != 0 ? 'r' : '-',
	     (st->st_mode & S_IWOTH) != 0 ? 'w' : '-',
	     (st->st_mode & S_ISVTX) != 0 ? 't' :
	     (st->st_mode & S_IXOTH) ? 'x' : '-',

	     st->st_nlink, /* field will be misaligned if > 99 */
	     owner,
	     group,
	     (long) st->st_size,
	     basename,
	     S_ISREG(st->st_mode) && (st->st_mode & S_IXUSR) ? "*" :
	     S_ISREG(st->st_mode) ? "" : /* 90% is of this type, so check now*/
	     S_ISDIR(st->st_mode) ? "/" :
	     S_ISLNK(st->st_mode) ? "@" :
	     S_ISFIFO(st->st_mode) ? "|" :
	     S_ISSOCK(st->st_mode) ? "=" :
	     /* no symbol for block and character devices */
	     "");

    new_entry = safe_malloc(sizeof(struct dir_list_entry));
    new_entry->menu_line = safe_strdup(menu_line);
    new_entry->basename = safe_strdup(basename);
    new_entry->lstat_st_mode = st->st_mode;
    new_entry->next = NULL;

    if (first_dir_list_entry == NULL) {
	first_dir_list_entry = last_dir_list_entry = new_entry;
	first_dir_list_entry->prev = NULL;
    } else {
	new_entry->prev = last_dir_list_entry;
	last_dir_list_entry->next = new_entry;
	last_dir_list_entry = new_entry;
    }
}

/*
 * Add a file to the end of the playlist.
 * Return 0 on success, otherwise tell the user about the error and return -1.
 */
int browser_add_song(int index)
{
    struct stat st;
    struct dir_list_entry *p;
    char path[PATH_MAX + 1];

    p = get_pointer_to_menu_entry(index);

    if (!S_ISREG(p->lstat_st_mode) && !S_ISLNK(p->lstat_st_mode)) {
	msg("%s is not a regular file.", p->basename);
	return -1;
    }

    /* See if the link points to a directory. */
    if (S_ISLNK(p->lstat_st_mode)) {
	if (stat(p->basename, &st) < 0) {
	    perror_msg("Can't stat %s", p->basename);
	    return -1;
	}
	if (!S_ISREG(st.st_mode)) {
	    msg("Symlink %s does not point to a directory.",
		     p->basename);
	    return -1;
	}
    }

    if (!has_registered_filename_extension(p->basename)) {
        MSG_FILENAME_EXTENSION_NOT_REGISTERED(p->basename);
	return -1;
    }

    /* add_song() wants absolute paths */
    snprintf(path, sizeof(path), "%s%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/",
	     p->basename);

    add_song(path);

    return 0;
}

/*
 * Ask for a directory to browse. If input is received, browse the directory.
 * This function is just about the same as main_chdir().
 */
void browser_chdir(struct menu *menu)
{
    char path[PATH_MAX + 1];
    struct stat st;

    snprintf(path, sizeof(path), "%s%s",
	     cwd(),
	     IS_ROOT_PATH(cwd()) ? "" : "/");

    if (enter_path(path, sizeof(path), DIRECTORY | MUST_EXIST | HALF_FILLED,
		   "Change directory to: ") == NULL) {
	return;
    }
    expand_path(path, sizeof(path));
    
    if (stat(path, &st) == 0) {
	if (S_ISDIR(st.st_mode)) {
	    int n;
	    n = browser_examine_directory(path);
	    if (n == -1) {
		n = browser_examine_directory(cwd());
		if (n == -1) {
		    n = browser_examine_directory(HomeDir);
		    if (n == -1) {
			return;
		    }
		}
	    }
	    menu->entries = n;
	    menu->oldcurrent = menu->current = menu->top = 0;
	    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	    TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
	} else
	    msg("%s is not a directory.", path);
    } else
	perror_msg("Can't stat %s", path);
}

/*
 * `path' must be an absolute path to a directory.
 * On success, return number of entries added, otherwise tell the user
 * about the error and return -1.
 */
int browser_examine_directory(const char *path)
{
    struct stat st;
    DIR *dp;
    struct dirent *de;
    int entries;
    char buf[PATH_MAX + SHORT_STRING];

    browser_free_dir_list();

    /* xchdir() and add_dir_entry() wants absolute paths. */
    assert(path[0] == '/');

    if (stat(path, &st) == -1) {
	perror_msg("Can't stat %s", path);
	return -1;
    }
    if (!S_ISDIR(st.st_mode)) {
	msg("%s is not a directory", path);
	return -1;
    }
    if ((dp = opendir(path)) == NULL) {
	perror_msg("Can't open directory %s", path);
	return -1;
    }
    /* When browsing the file system, we always change directory to the
       directory the user sees in the browser. */
    if (xchdir(path) != 0) {
	perror_msg("Can't change directory to %s", path);
	return -1;
    }

    entries = 0;
    while ((de = readdir(dp)) != NULL) {
	snprintf(buf, sizeof(buf), "%s%s%s",
		 path, path[strlen(path) - 1] == '/' ? "" : "/", de->d_name);
	/* Ignore stat errors. */
	if (lstat(buf, &st) == -1)
	    continue;
	add_dir_entry(de->d_name, &st);
	entries++;
    }

    browser_sort_dir_entries();
    closedir(dp);

    return entries;
}

/*
 * Free all memory allocated to the current directory listing, then set
 * `first_dir_list_entry' and `last_dir_list_entry` to NULL.
 */
void browser_free_dir_list(void)
{
    struct dir_list_entry *p, *tmp;

    for (p = first_dir_list_entry; p; p = tmp) {
	tmp = p->next;
	safe_free(p->menu_line);
	safe_free(p->basename);
	safe_free(p);
    }

    first_dir_list_entry = last_dir_list_entry = NULL;
}

/*
 * Return pointer to entry number `index' in the directory listing.
 */
static struct dir_list_entry *get_pointer_to_menu_entry(int index)
{
    struct dir_list_entry *p = first_dir_list_entry;
    int i = 0;

    while (i++ < index)
	p = p->next;

    return p;
}

/*
 * Copy line to display in the browser menu for entry number `index' in
 * the directory listing into `str'.
 */
void browser_make_menu_entry(char *str, size_t len, int index)
{
    int i = 0;
    struct dir_list_entry *p = first_dir_list_entry;

    if (index != 0)
	while (i++ < index)
	    p = p->next;

    snprintf(str, len, "%s", p->menu_line);
}

/*
 * Browse the highlighted directory in the directory listing.
 */
void browser_open_directory(struct menu *menu)
{
    int i;
    char dir[PATH_MAX + 1];
    struct stat st;
    struct dir_list_entry *p;

    p = get_pointer_to_menu_entry(menu->current);

    if (!S_ISDIR(p->lstat_st_mode) && !S_ISLNK(p->lstat_st_mode)) {
	msg("%s is not a directory.");
	return;
    }

    if (S_ISLNK(p->lstat_st_mode)) {
	if (stat(p->basename, &st) < 0) {
	    perror_msg("Can't stat %s", p->basename);
	    return;
	}
	if (!S_ISDIR(st.st_mode)) {
	    msg("Symlink %s does not point to a directory.",
		     p->basename);
	    return;
	}
    }

    /* If parent directory */
    if (strcmp(p->basename, "..") == 0) {
	char *cp;

	if (IS_ROOT_PATH(cwd())) {
	    msg("You are in the root directory.");
	    return;
	}

	strfcpy(dir, cwd(), sizeof(dir));
	cp = strrchr(dir + 1, '/');
	if (cp)
	    *cp = '\0'; /* remove trailing slash */
	else
	    dir[1] = '\0'; /* parent directory is the root directory */
      
    } else { /* Not parent directory */
	snprintf(dir, sizeof(dir), "%s%s%s",
		 cwd(),
		 IS_ROOT_PATH(cwd()) ? "" : "/",
		 p->basename);
    }
    
    i = browser_examine_directory(dir);
    if (i == -1) {
	/* try to restore the old values */
	i = browser_examine_directory(cwd());
	if (i == -1) {
	    i = browser_examine_directory(cwd());
	    if (i == -1)
		return;
	}
    }
    menu->entries = i;
    menu->oldcurrent = menu->current = menu->top = 0;
	    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
	    TURN_ON_FLAG(menu->redraw, REDRAW_STATUS);
}

/*
 * Sort directory listing in decreasing/reverse order.
 * This is exactly the same function as browse_sort_dir_entries() except
 * that GT is replaced with LT and LT with GT.
 */
void browser_reverse_sort_dir_entries(void)
{
    /* the first unsorted node to be inserted */
    struct dir_list_entry *fu;
    /* the last sorted node (tail of sorted sublist) */
    struct dir_list_entry *ls;
    struct dir_list_entry *cur, *trailing;

    if (!first_dir_list_entry)
	return;

    ls = first_dir_list_entry;	/* An empty list is already sorted. */
    while (ls->next) {
	fu = ls->next;		/* Remember first unsorted node. */

	if (GT(fu->basename, first_dir_list_entry->basename)) {
	    ls->next = fu->next;
	    fu->next = first_dir_list_entry;
	    fu->prev = NULL;
	    first_dir_list_entry->prev = fu;
	    /* Insert first unsorted at the head of sorted list. */
	    first_dir_list_entry = fu;
	} else {		/* Search the sorted sublist. */
	    trailing = first_dir_list_entry;
	    for (cur = trailing->next;
		 LT(fu->basename, cur->basename);
		 cur = cur->next)
		{
		    trailing = cur;
		}

	    if (fu == cur)
		ls = fu;
	    else {
		/* First unsorted node now belongs between trailing and cur. */
		ls->next = fu->next;
		if (fu->next)
		    fu->next->prev = ls;
		fu->next = cur;
		cur->prev = fu;
		fu->prev = trailing;
		trailing->next = fu;
	    }
	}
    }
}

/*
 * Do a regular expression search through the basename of each entry in
 * the directory listing until a match is found or all entries have
 * been searched. If the former happens, update `menu' so the matching entry
 * will be highlighted on next redraw.
 */
void browser_search(struct menu *menu, int op)
{
    /* current directory listing entry being compared to the regexp */
    struct dir_list_entry *cur;
    /* index position of current directory listing entry being compared to
       the regexp */
    int cur_index_pos;
    char search_str[32];     /* search string */
    bool first_time_through; /* used by the search loop */
    regex_t preg;            /* compiled regexp structure, must be regfree()d*/
    int errcode;             /* error code to check for regcomp() error */

    /* make `cur' point to the current entry in the directory listing and
       `cur_index_pos' contain its position/index number */
    cur = first_dir_list_entry;
    cur_index_pos = 0;
    while (cur_index_pos++ < menu->current)
	cur = cur->next;
    cur_index_pos--;

    if (op == CMD_REPEAT_LAST_SEARCH || op == CMD_REPEAT_LAST_SEARCH_OPPOSITE) {
        if (menu->prev_main_search_op == NOP) {
	    msg("No previous searches.");
	    return;
	}
	strfcpy(search_str, menu->prev_search_str, sizeof(search_str));
	if (op == CMD_REPEAT_LAST_SEARCH_OPPOSITE) {
	    /* opposite direction of previous main search operation */
	    if (menu->prev_main_search_op == CMD_REGEXP_SEARCH_BACKWARD)
		op = CMD_REGEXP_SEARCH_FORWARD;
	    else
		op = CMD_REGEXP_SEARCH_BACKWARD;
	} else { /* op == CMD_REPEAT_LAST_SEARCH */
    	    op = menu->prev_main_search_op;
	}
    } else if (op == CMD_REGEXP_SEARCH_BACKWARD || op == CMD_REGEXP_SEARCH_FORWARD) {
	if (enter_string(search_str, sizeof(search_str),
			 EMPTY, NO_BROWSE, "%s", op == CMD_REGEXP_SEARCH_FORWARD ?
			 "Search for: " :
			 "Reverse search for: ") == NULL) {
	    return;
	}
	strfcpy(menu->prev_search_str, search_str,
		sizeof(menu->prev_search_str));
	menu->prev_main_search_op = op;
    } else /* unknown search operation */
	assert(0);

    /* start search from the next or previous line */
    if (op == CMD_REGEXP_SEARCH_FORWARD) {
	cur = cur->next;
	cur_index_pos++;
    } else if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	cur = cur->prev;
	cur_index_pos--;
    }

    /* compile `search_str' into a regular expression */
    if (is_boolean_option_true(OPT_IGNORE_CASE))
	errcode = regcomp(&preg, search_str,
			  REG_NOSUB | REG_ICASE | REG_EXTENDED);
    else /* do not ignore case */
	errcode = regcomp(&preg, search_str, REG_NOSUB | REG_EXTENDED);

    if (errcode) {
	display_regexp_error(errcode, &preg);
	return;
    }

    first_time_through = TRUE;
 loop_again:
    while (cur) {

	/* does `cur->basename' match the regexp `preg'? */
	if (regexec(&preg, cur->basename, -1, NULL, 0) == 0)
	    goto found;

	if (op == CMD_REGEXP_SEARCH_FORWARD) {
	    if (cur->next)
		cur_index_pos++;
	    cur = cur->next;
	} else if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	    if (cur->prev)
		cur_index_pos--;
	    cur = cur->prev;
	} else
	    assert(0);
    }

    /* if we're on the first time through, we wrap the search */
    if (first_time_through == TRUE) {
	first_time_through = FALSE;
	cur = first_dir_list_entry;
	cur_index_pos = 0;
	if (op == CMD_REGEXP_SEARCH_BACKWARD) {
	    while (cur->next) {
		cur_index_pos++;
		cur = cur->next;
	    }
	} /* else op == CMD_REGEXP_SEARCH_FORWARD, so don't do anything since
	     `cur' points to the first entry and `cur_index_pos' is 0 */

	goto loop_again;
    }
    msg("Not found.");
    regfree(&preg); /* `preg' contains pointers to allocated memory blocks */
    return;

 found:
    /* if search was wrapped */
    if (first_time_through == FALSE) {
	msg("Search hit %s.",
		    op == CMD_REGEXP_SEARCH_FORWARD ?
		    "bottom, continuing at top" :	/* forward search */
		    "top, continuing at bottom");	/* backward search */
    }
    if (cur_index_pos != menu->current) {
	menu->current = cur_index_pos;
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
    }
    regfree(&preg); /* `preg' contains pointers to allocated memory blocks */
}

/*
 * Sort the directory listing in increasing order.
 */
void browser_sort_dir_entries(void)
{
    /* the first unsorted node to be inserted */
    struct dir_list_entry *fu;
    /* the last sorted node (tail of sorted sublist) */
    struct dir_list_entry *ls;
    struct dir_list_entry *cur, *trailing;

    if (!first_dir_list_entry)
	return;

    ls = first_dir_list_entry;	/* An empty list is already sorted. */
    while (ls->next) {
	fu = ls->next;		/* Remember first unsorted node. */

	if (LT(fu->basename, first_dir_list_entry->basename)) {
	    ls->next = fu->next;
	    fu->next = first_dir_list_entry;
	    first_dir_list_entry->prev = fu;
	    /* Insert first unsorted at the head of sorted list. */
	    first_dir_list_entry = fu;
	} else {		/* Search the sorted sublist. */
	    trailing = first_dir_list_entry;
	    for (cur = trailing->next;
		 GT(fu->basename, cur->basename);
		 cur = cur->next)
		{
		    trailing = cur;
		}

	    if (fu == cur)
		ls = fu;
	    else {
		/* First unsorted node now belongs between trailing and cur. */
		ls->next = fu->next;
		if (fu->next)
		    fu->next->prev = ls;
		fu->next = cur;
		cur->prev = fu;
		fu->prev = trailing;
		trailing->next = fu;
	    }
	}
    }
}
